/*
 * tcpif.cpp
 *
 *
 */



#include <tcpif.h>
#include <lwip/tcp.h>
#include <cocoos.h>
#include <stdint.h>

#include "lwip/debug.h"
#include "lwip/stats.h"
#include <inc/pin.h>
#include <tcp_server.h>
#include <tcp_defs.h>
#include <debug.h>


static tcpPacket packetBuf[128];
tcpPacketQueue tcpif::tq(packetBuf, 128);

static tcpif *inst;
extern Evt_t tcpEvent;

tcpif::tcpif() {
    inst = this;
}

tcpif::~tcpif() {}

void tcpif::create(uint8_t prio) {
    task_create( task, prio, 0, 0, 0 );
}

// Waits for the tcpEvent and pops and processes a tcpPacket from the queue.
void tcpif::task() {
    static tcpPacket msg;
    task_open();
    for (;;) {
        event_wait(tcpEvent);
        while (inst->checkMessage(&msg)) {
            debug_print("tcpif: popped a packet: \n");
            if (msg.type == TCP_RECV_MSG) {
                debug_print("TCP_RECV_MSG\n");
                tcp_server_receive(&msg);
            }
            else if (msg.type == TCP_ERROR_MSG) {
                debug_print("TCP_ERROR_MSG\n");
                tcp_server_errormsg(&msg);
            }
            else if (msg.type == TCP_SEND_MSG) {
                debug_print("TCP_SEND_MSG\n");
                tcp_server_send(msg.tpcb, (struct tcp_server_struct *)msg.arg);
            }
            else if (msg.type == TCP_CLOSE_MSG) {
                debug_print("TCP_CLOSE_MSG\n");
                tcp_server_close(msg.tpcb, (struct tcp_server_struct *)msg.arg);
            }
            else {
                debug_print("unknown message\n");
            }
            debug_print("tcpif: packet done\n");
        }
    }
    task_close();
}

uint8_t tcpif::init() {
    struct tcp_pcb *tcp_server_pcb;
    uint8_t result = 0;

    /* create new tcp pcb */
    tcp_server_pcb = tcp_new();

    if (tcp_server_pcb != NULL) {
        err_t err;

        /* bind echo_pcb to port 7 (ECHO protocol) */
        err = tcp_bind(tcp_server_pcb, IP_ADDR_ANY, 7);

        if (err == ERR_OK) {
            /* start tcp listening for echo_pcb */
            tcp_server_pcb = tcp_listen(tcp_server_pcb);

            /* initialize LwIP tcp_accept callback function */
            tcp_accept(tcp_server_pcb, accept);
            result = 1;
        }
    }
    return result;
}

bool tcpif::checkMessage(tcpPacket *msg) {
    return tq.pop(msg);
}

//
// TCP event callbacks, called from within the context of Lwipif::handlePacket()
// Saves the connection info pointers into a tcpPacket and pushes it to the queue, and
// signals an event to indicate a queue change.
//

err_t tcpif::accept(void *arg, struct tcp_pcb *newpcb, err_t err) {
    err_t ret_err;
    struct tcp_server_struct *es;

    /* set priority for the newly accepted tcp connection newpcb */
    tcp_setprio(newpcb, TCP_PRIO_MIN);

    /* allocate structure es to maintain tcp connection informations */
    es = (struct tcp_server_struct *)mem_malloc(sizeof(struct tcp_server_struct));

    if (es != NULL) {
        es->state = ES_ACCEPTED;
        es->pcb = newpcb;
        es->p = NULL;

        debug_print("tcpif: accept\n");
        /* pass newly allocated es structure as argument to newpcb */
        tcp_arg(newpcb, es);

        /* initialize lwip tcp_recv callback function for newpcb  */
        tcp_recv(newpcb, recv);

        /* initialize lwip tcp_err callback function for newpcb  */
        tcp_err(newpcb, error);

        /* initialize lwip tcp_poll callback function for newpcb */
        tcp_poll(newpcb, poll, 1);

        ret_err = ERR_OK;
    }
    else {
        /* return memory error */
        ret_err = ERR_MEM;
    }

    return ret_err;
}

err_t tcpif::recv(void *arg, struct tcp_pcb *tpcb, struct pbuf *p, err_t err) {
    tcpPacket msg;
    msg.type = TCP_RECV_MSG;
    msg.arg = arg;
    msg.tpcb = tpcb;
    msg.p = p;
    msg.err = err;

    tq.push(msg);
    event_ISR_signal(tcpEvent);
    debug_print("tcpif: recv\n");
    return ERR_OK;
}

void tcpif::error(void *arg, err_t err) {
    tcpPacket msg;
    msg.type = TCP_ERROR_MSG;
    msg.arg = arg;
    msg.err = err;

    tq.push(msg);
    event_ISR_signal(tcpEvent);
    debug_print("tcpif: error\n");
}

err_t tcpif::poll(void *arg, struct tcp_pcb *tpcb) {
    tcpPacket msg;
    err_t ret_err;

    struct tcp_server_struct *es;

    es = (struct tcp_server_struct *)arg;
    debug_print("tcpif: poll\n");
    if (es != NULL) {
        if (es->p != NULL) {
            /* there is a remaining pbuf (chain) , try to send data */
            msg.type = TCP_SEND_MSG;
            msg.arg = arg;
            msg.tpcb = tpcb;
            tq.push(msg);
            event_ISR_signal(tcpEvent);
        }
        else {
            /* no remaining pbuf (chain)  */
            if (es->state == ES_CLOSING) {
                /*  close tcp connection */
                msg.type = TCP_CLOSE_MSG;
                msg.arg = arg;
                msg.tpcb = tpcb;
                tq.push(msg);
                event_ISR_signal(tcpEvent);
            }
        }
        ret_err = ERR_OK;
    }
    else {
        /* nothing to be done */
        tcp_abort(tpcb);
        ret_err = ERR_ABRT;
    }

    return ret_err;
}




